@@ -1,15 +1,41 @@
-import errno
+import os
 import socket
 from itertools import chain, imap
-from redis.exceptions import ConnectionError, ResponseError, InvalidResponse
+from redis.exceptions import (
+    RedisError,
+    ConnectionError,
+    ResponseError,
+    InvalidResponse,
+    AuthenticationError
+)
+
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from StringIO import StringIO
+
+try:
+    import hiredis
+    hiredis_available = True
+except ImportError:
+    hiredis_available = False
 
 class PythonParser(object):
+    "Plain Python parsing class"
+    MAX_READ_LENGTH = 1000000
+
     def __init__(self):
         self._fp = None
 
+    def __del__(self):
+        try:
+            self.on_disconnect()
+        except:
+            pass
+
     def on_connect(self, connection):
         "Called when the socket connects"
-        self._fp = connection._sock.makefile('r')
+        self._fp = connection._sock.makefile('rb')
 
     def on_disconnect(self):
         "Called when the socket disconnects"
@@ -24,7 +50,25 @@ class PythonParser(object):
         """
         try:
             if length is not None:
-                return self._fp.read(length+2)[:-2]
+                bytes_left = length + 2 # read the line ending
+                if length > self.MAX_READ_LENGTH:
+                    # apparently reading more than 1MB or so from a windows
+                    # socket can cause MemoryErrors. See:
+                    # https://github.com/andymccurdy/redis-py/issues/205
+                    # read smaller chunks at a time to work around this
+                    try:
+                        buf = StringIO()
+                        while bytes_left > 0:
+                            read_len = min(bytes_left, self.MAX_READ_LENGTH)
+                            buf.write(self._fp.read(read_len))
+                            bytes_left -= read_len
+                        buf.seek(0)
+                        return buf.read(length)
+                    finally:
+                        buf.close()
+                return self._fp.read(bytes_left)[:-2]
+
+            # no length, read a full line
             return self._fp.readline()[:-2]
         except (socket.error, socket.timeout), e:
             raise ConnectionError("Error while reading from socket: %s" % \
@@ -68,6 +112,17 @@ class PythonParser(object):
         raise InvalidResponse("Protocol Error")
 
 class HiredisParser(object):
+    "Parser class for connections using Hiredis"
+    def __init__(self):
+        if not hiredis_available:
+            raise RedisError("Hiredis is not installed")
+
+    def __del__(self):
+        try:
+            self.on_disconnect()
+        except:
+            pass
+
     def on_connect(self, connection):
         self._sock = connection._sock
         self._reader = hiredis.Reader(
@@ -79,6 +134,8 @@ class HiredisParser(object):
         self._reader = None
 
     def read_response(self):
+        if not self._reader:
+            raise ConnectionError("Socket closed on remote end")
         response = self._reader.gets()
         while response is False:
             try:
@@ -96,17 +153,18 @@ class HiredisParser(object):
             response = self._reader.gets()
         return response
 
-try:
-    import hiredis
+if hiredis_available:
     DefaultParser = HiredisParser
-except ImportError:
+else:
     DefaultParser = PythonParser
 
+
 class Connection(object):
     "Manages TCP communication to and from a Redis server"
     def __init__(self, host='localhost', port=6379, db=0, password=None,
                  socket_timeout=None, encoding='utf-8',
                  encoding_errors='strict', parser_class=DefaultParser):
+        self.pid = os.getpid()
         self.host = host
         self.port = port
         self.db = db
@@ -117,6 +175,12 @@ class Connection(object):
         self._sock = None
         self._parser = parser_class()
 
+    def __del__(self):
+        try:
+            self.disconnect()
+        except:
+            pass
+
     def connect(self):
         "Connects to the Redis server if not already connected"
         if self._sock:
@@ -146,7 +210,6 @@ class Connection(object):
             return "Error %s connecting %s:%s. %s." % \
                 (exception.args[0], self.host, self.port, exception.args[1])
 
-
     def on_connect(self):
         "Initialize the connection, authenticate and select a database"
         self._parser.on_connect(self)
@@ -155,7 +218,7 @@ class Connection(object):
         if self.password:
             self.send_command('AUTH', self.password)
             if self.read_response() != 'OK':
-                raise ConnectionError('Invalid Password')
+                raise AuthenticationError('Invalid Password')
 
         # if a database is specified, switch to it
         if self.db:
@@ -223,6 +286,7 @@ class UnixDomainSocketConnection(Connection):
     def __init__(self, path='', db=0, password=None,
                  socket_timeout=None, encoding='utf-8',
                  encoding_errors='strict', parser_class=DefaultParser):
+        self.pid = os.getpid()
         self.path = path
         self.db = db
         self.password = password
@@ -255,6 +319,7 @@ class ConnectionPool(object):
     "Generic connection pool"
     def __init__(self, connection_class=Connection, max_connections=None,
                  **connection_kwargs):
+        self.pid = os.getpid()
         self.connection_class = connection_class
         self.connection_kwargs = connection_kwargs
         self.max_connections = max_connections or 2**31
@@ -262,8 +327,14 @@ class ConnectionPool(object):
         self._available_connections = []
         self._in_use_connections = set()
 
+    def _checkpid(self):
+        if self.pid != os.getpid():
+            self.disconnect()
+            self.__init__(self.connection_class, self.max_connections, **self.connection_kwargs)
+
     def get_connection(self, command_name, *keys, **options):
         "Get a connection from the pool"
+        self._checkpid()
         try:
             connection = self._available_connections.pop()
         except IndexError:
@@ -280,8 +351,10 @@ class ConnectionPool(object):
 
     def release(self, connection):
         "Releases the connection back to the pool"
-        self._in_use_connections.remove(connection)
-        self._available_connections.append(connection)
+        self._checkpid()
+        if connection.pid == self.pid:
+            self._in_use_connections.remove(connection)
+            self._available_connections.append(connection)
 
     def disconnect(self):
         "Disconnects all connections in the pool"
